#include <dirent.h>
#include <fcntl.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <utime.h>
#include <stdlib.h>
#include <stdio.h>

#include "update_tree.h"


#define BUFSIZE 1000

bool copy_file(int source_fd, int target_fd) {

    // copy file content
    char*  buf = malloc(BUFSIZE);
    int nbytes;
    while (true) {
        // fill buffer 
        if ((nbytes = read(source_fd, buf, BUFSIZE)) < 0) {
            perror("Copying failed while reading");
            free(buf);
            return false;
        }
        // stop if EOF is reached
        if (nbytes == 0) break;
        // write buffer
        if (write(target_fd, buf, nbytes) != nbytes) {
            perror("Copying failed while writing");
            free(buf);
            return false;
        }
    }
    free(buf);
    
    // get stat from input file
    struct stat stat_buf;
    if (fstat(source_fd, &stat_buf) < 0) {
        return false;
    }

    // copy file protection (mode)
    if (fchmod(target_fd, stat_buf.st_mode) < 0) {
        return false;
    }

    // copy atime & mtime
    struct timespec times[2];
    times[0] = stat_buf.st_atim;
    times[1] = stat_buf.st_mtim;
    if (futimens(target_fd, times) < 0) {
        return false;
    }
     
    close(source_fd);
    if (close(target_fd) < 0) return false;
    return true;
}

bool update_tree(int source_dir_fd, int target_dir_fd) {
    // open directory pointer from file descriptor
    DIR* dir = fdopendir(source_dir_fd);
    if (!dir) return false;
    struct dirent* entry;
    
    // go through all entries of the directory
    while ((entry = readdir(dir))) {
        // skip entries "." and ".."
        if (strcmp(entry->d_name, ".") == 0) continue;
        if (strcmp(entry->d_name, "..") == 0) continue;

        // open entry, which is next source file
        int source_fd = openat(source_dir_fd, entry->d_name, O_RDONLY);
        if (source_fd < 0) {
            return false;
        }
        
        // obtain information from source
        struct stat statbuf_source;
        if (fstat(source_fd, &statbuf_source) < 0) {
            return false;
        }
        bool source_isreg = S_ISREG(statbuf_source.st_mode);
        bool source_isdir = S_ISDIR(statbuf_source.st_mode);
        
        int oflag;
        if (source_isreg) oflag = O_RDWR;
        else if (source_isdir) oflag = O_RDONLY;
        else continue; // skip handling of source if it is neither a regular file nor a directory

        // try to open target
        int target_fd = openat(target_dir_fd, entry->d_name, oflag);
        if (target_fd < 0) { // target does not exist
            if (source_isreg) {
               target_fd = openat(target_dir_fd, entry->d_name,
                                  O_RDWR|O_CREAT|O_TRUNC, 0600);
            } else { // source is directory
                static int mkdir_count = 0;
                if (mkdir_count++ > 100 || mkdirat(target_dir_fd, entry->d_name, 0777) < 0) {
                    return false;
                }
                target_fd = openat(target_dir_fd, entry->d_name, O_RDONLY);
            }
            if (target_fd < 0) {
                return false;
            }
        }
        
        // obtain information from target
        struct stat statbuf_target;
        if (fstat(target_fd, &statbuf_target) < 0) return false;
        bool target_isreg = S_ISREG(statbuf_target.st_mode);
        bool target_isdir = S_ISDIR(statbuf_target.st_mode);
        
        // skip update if target has other type than directory or regular file ...
        if (!target_isreg && !target_isdir) continue;
        // ... or if source and target have different types
        if (source_isreg != target_isreg) continue;
        
        if (source_isdir) {
            if (!update_tree(source_fd, target_fd)) return false;
        } 
        else {
            if (!copy_file(source_fd, target_fd)) return false;
        }
    }
    closedir(dir);
    return true;
}
